import * as utility from "./toolUtil.js";
let filterResponse = {};
let isFilterResponseCalculated = false;

// ---------- Controls & constants ----------
export function updateFilterType(args) {
  let configData = args.parameterConfigData[args.selectedDevice];
  if (configData.filterType) {
    let filterTypeObject = get_Control_Object(configData.filterType);
    return {
      options: filterTypeObject.options,
      value: filterTypeObject.value,
    };
  }
  return {};
}

export function updateClockFrequency(args) {
  let configData = args.parameterConfigData[args.selectedDevice];
  if (configData.clockFrequency) {
    let valueArray = configData.clockFrequency.split("<");
    return {
      value: parseFloat(valueArray[1]),
      minValue: parseFloat(
        valueArray
          .find(element => element.includes("(min)"))
          .replace("(min)", "")
          .trim(),
      ),
      maxValue: parseFloat(
        valueArray
          .find(element => element.includes("(max)"))
          .replace("(max)", "")
          .trim(),
      ),
    };
  }
  return {};
}

export function updateDataRate(args) {
  let configData = args.parameterConfigData[args.selectedDevice];
  if (configData.dataRate) {
    let dataRateObject = get_Control_Object(configData.dataRate);
    return {
      options: dataRateObject.options,
      value: dataRateObject.value,
    };
  }
  return {};
}

export function updateProgrammableDelay(args) {
  let configData = args.parameterConfigData[args.selectedDevice];
  if (configData.programmableDelay) {
    let programmableDelayObject = get_Control_Object(
      configData.programmableDelay,
    );
    return {
      options: programmableDelayObject.options,
      value: programmableDelayObject.value,
    };
  }
  return {};
}

export function updateGlobalChop(args) {
  let configData = args.parameterConfigData[args.selectedDevice];
  if (configData.globalChop) {
    let globalChopObject = get_Control_Object(configData.globalChop);
    return Object.keys(globalChopObject.options).length === 2
      ? {
          value: globalChopObject.value === 0 ? false : true,
          labelForUnchecked: globalChopObject.options[0],
          labelForChecked: globalChopObject.options[1],
        }
      : {
          controlType: "input-select",
          options: globalChopObject.options,
          value: globalChopObject.value,
        };
  }
  return {};
}

// ---------- Helper Methods ---------------

function get_Control_Object(controlData) {
  // Split the string by pipe characters (|)
  const valueArray = controlData.split("|");

  // Find the index of the element containing "(default)"
  const defaultValueIndex = valueArray.findIndex(element =>
    element.includes("(default)"),
  );

  // Map the array values to an object with indices
  const controlObject = {
    options: valueArray.reduce((acc, value, index) => {
      acc[index] = value.replace("(default)", "").trim();
      return acc;
    }, {}),
    value: defaultValueIndex !== -1 ? defaultValueIndex : 0,
  };

  return controlObject;
}

export function computeConversionTime(args) {
  let table = [];
  let timeTo1stConversion;
  let conversionPeriod;
  const filter = args.filterType.value;
  const dataRateSetting = args.dataRateSetting.value;
  const chop =
    typeof args.globalChop.value === "boolean"
      ? Number(args.globalChop.value)
      : args.globalChop.value;
  const fClk_MHz = args.clockFrequencyMHz.value;
  const delay = args.programmableDelay.value;
  let module = args.parameterConfigData[args.selectedDevice].calculationModule;
  if (Object.keys(module).length > 0) {
    timeTo1stConversion = module
      .getFilterSettlingTime(
        filter,
        dataRateSetting,
        fClk_MHz * 1000000,
        chop,
        delay,
      )
      .toFixed(2);
    conversionPeriod = module.deviceConfig.single_conversion_time_ms.toFixed(2);
  } else {
    const programmableDelay = parseInt(
      args.delay_list.value[args.programmableDelay.value],
    );
    const sinc3_OSR = args.data_rate_osr_value.value[dataRateSetting];
    const filterType = args.filterType.options[args.filterType.value];
    const chopEnabled = args.globalChop.value;
    const sinc1_OSR_List = args.sinc1_data_rate_osr_value.value;
    const FIRST_DATA =
      args.first_data.value[args.filterType.value][dataRateSetting];
    let osrList = getOSRList(
      filterType,
      dataRateSetting,
      sinc3_OSR,
      sinc1_OSR_List,
    );
    const totalOSR = osrList.reduce(
      (accumulator, currentValue) => accumulator * currentValue,
    );

    timeTo1stConversion = this.getTimeTo1stConversion(
      fClk_MHz,
      programmableDelay,
      FIRST_DATA,
      chopEnabled,
    ).toFixed(2);
    conversionPeriod = this.getConversionPeriod(
      fClk_MHz,
      totalOSR,
      timeTo1stConversion,
      chopEnabled,
    ).toFixed(2);
  }
  table.push(
    utility.getTableRow(
      [
        {
          value: "Conversion Time",
          colspan: 2,
        },
      ],
      "head",
    ),
  );

  table.push(
    utility.getTableRow(
      [
        {
          value: "Time to 1st Conversion",
        },
        {
          value: "Conversion Period",
        },
      ],
      "head",
    ),
  );

  table.push(
    utility.getTableRow([
      {
        value: `${timeTo1stConversion} ms`,
      },
      {
        value: `${conversionPeriod} ms`,
      },
    ]),
  );

  return {
    value: table,
  };
}

export function computeFilterResponseGraph(args) {
  let adjFreqArr = filterResponse.adjFreqArr;
  let effectiveNoiseBandwidth = filterResponse.effectiveNoiseBandwidth;
  let magnitudeArr = filterResponse.magnitudeArr;
  let normalModeRangeDetails = filterResponse.normalModeRangeDetails;
  let effectiveDataRate = filterResponse.effectiveDataRate;
  let minus3DbCutoffFreq = filterResponse.minus3DbCutoffFreq;
  let fclk_nom = filterResponse.fclk_nom;
  const fClk_MHz = args.clockFrequencyMHz.value;
  if (!isFilterResponseCalculated) {
    const filter = args.filterType.value;
    const dataRateSetting = args.dataRateSetting.value;
    const chop =
      typeof args.globalChop.value === "boolean"
        ? Number(args.globalChop.value)
        : args.globalChop.value;
    const frequencyInp = args.frequency.value;
    const toleranceInp = args.tolerance.value;
    const lutData = args.parameterConfigData[args.selectedDevice].lutData;
    let module =
      args.parameterConfigData[args.selectedDevice].calculationModule;
    if (Object.keys(module).length > 0) {
      magnitudeArr = module.getFilterResponse(
        filter,
        dataRateSetting,
        fClk_MHz * 1000000,
        chop,
        lutData,
      );
      effectiveDataRate = module.deviceConfig.data_rate_sps;
      adjFreqArr = module.deviceConfig.x_freqs_adj;
      fclk_nom = module.deviceConfig.fclk_nom / 1000000;
      ({ effectiveNoiseBandwidth, minus3DbCutoffFreq, normalModeRangeDetails } =
        this.getFilterResponseOutputs(
          adjFreqArr,
          magnitudeArr,
          frequencyInp,
          toleranceInp,
        ));
      filterResponse.effectiveNoiseBandwidth = effectiveNoiseBandwidth;
      filterResponse.minus3DbCutoffFreq = minus3DbCutoffFreq;
      filterResponse.normalModeRangeDetails = normalModeRangeDetails;
      filterResponse.effectiveDataRate = effectiveDataRate;
      isFilterResponseCalculated = true;
    } else {
      fclk_nom = args.nominal_fClk_MHz.value;
      const sinc3_OSR = args.data_rate_osr_value.value[dataRateSetting];
      const filterType = args.filterType.options[args.filterType.value];
      const chopEnabled = args.globalChop.value;
      const sinc1_OSR_List = args.sinc1_data_rate_osr_value.value;
      ({
        adjFreqArr,
        effectiveNoiseBandwidth,
        magnitudeArr,
        minus3DbCutoffFreq,
        normalModeRangeDetails,
        effectiveDataRate,
      } = this.getFilterResponseDetails(
        filterType,
        chopEnabled,
        sinc3_OSR,
        sinc1_OSR_List,
        dataRateSetting,
        lutData,
        fClk_MHz,
        fclk_nom,
        frequencyInp,
        toleranceInp,
      ));
      filterResponse.effectiveNoiseBandwidth = effectiveNoiseBandwidth;
      filterResponse.minus3DbCutoffFreq = minus3DbCutoffFreq;
      filterResponse.normalModeRangeDetails = normalModeRangeDetails;
      filterResponse.effectiveDataRate = effectiveDataRate;
      isFilterResponseCalculated = true;
    }
  } else {
    isFilterResponseCalculated = false;
  }
  return {
    limits: {
      /**
       * Frequency array is calculated till fMod/2.
       *
       * In the graph, after mirroring, the limit will be till fMod.
       *
       * So, X axis max value => FMOD/2 * 2 => FMOD.
       */
      xaxisMax: adjFreqArr.at(-1) * 2,
      xaxisMin: 0,
      yaxisMax: 0,
      yaxisMin: -200,
    },
    metaData: {
      adjFreqToFreq: this.getFreqFromAdjustedFreqFunc(fClk_MHz, fclk_nom),
      filterResponse: {
        adjFreqArr,
        magnitudeArr,
      },
      normalModeRangeFreq: {
        min: normalModeRangeDetails.min,
        max: normalModeRangeDetails.max,
      },
      effectiveNoiseBandwidth,
      odr: effectiveDataRate,
    },
  };
}

export function computeFilterResponseOutputs(args) {
  let adjFreqArr = filterResponse.adjFreqArr;
  let magnitudeArr = filterResponse.magnitudeArr;
  let effectiveNoiseBandwidth = filterResponse.effectiveNoiseBandwidth;
  let minus3DbCutoffFreq = filterResponse.minus3DbCutoffFreq;
  let normalModeRangeDetails = filterResponse.normalModeRangeDetails;
  let effectiveDataRate = filterResponse.effectiveDataRate;
  if (!isFilterResponseCalculated) {
    const filter = args.filterType.value;
    const dataRateSetting = args.dataRateSetting.value;
    const fClk_MHz = args.clockFrequencyMHz.value;
    const chop =
      typeof args.globalChop.value === "boolean"
        ? Number(args.globalChop.value)
        : args.globalChop.value;
    const frequencyInp = args.frequency.value;
    const lutData = args.parameterConfigData[args.selectedDevice].lutData;
    const toleranceInp = args.tolerance.value;
    let module =
      args.parameterConfigData[args.selectedDevice].calculationModule;
    if (Object.keys(module).length > 0) {
      let magnitudeArr = module.getFilterResponse(
        filter,
        dataRateSetting,
        fClk_MHz * 1000000,
        chop,
        lutData,
      );
      let adjFreqArr = module.deviceConfig.x_freqs_adj;
      effectiveDataRate = module.deviceConfig.data_rate_sps;
      ({ effectiveNoiseBandwidth, minus3DbCutoffFreq, normalModeRangeDetails } =
        this.getFilterResponseOutputs(
          adjFreqArr,
          magnitudeArr,
          frequencyInp,
          toleranceInp,
        ));
      filterResponse.effectiveDataRate = effectiveDataRate;
      filterResponse.adjFreqArr = adjFreqArr;
      filterResponse.effectiveNoiseBandwidth = effectiveNoiseBandwidth;
      filterResponse.minus3DbCutoffFreq = minus3DbCutoffFreq;
      filterResponse.magnitudeArr = magnitudeArr;
      filterResponse.normalModeRangeDetails = normalModeRangeDetails;
      filterResponse.fclk_nom = module.deviceConfig.fclk_nom / 1000000;
      isFilterResponseCalculated = true;
    } else {
      const fclk_nom = args.nominal_fClk_MHz.value;
      const sinc3_OSR = args.data_rate_osr_value.value[dataRateSetting];
      const filterType = args.filterType.options[args.filterType.value];
      const chopEnabled = args.globalChop.value;
      const sinc1_OSR_List = args.sinc1_data_rate_osr_value.value;

      ({
        adjFreqArr,
        effectiveNoiseBandwidth,
        magnitudeArr,
        minus3DbCutoffFreq,
        normalModeRangeDetails,
        effectiveDataRate,
      } = this.getFilterResponseDetails(
        filterType,
        chopEnabled,
        sinc3_OSR,
        sinc1_OSR_List,
        dataRateSetting,
        lutData,
        fClk_MHz,
        fclk_nom,
        frequencyInp,
        toleranceInp,
      ));
      filterResponse.adjFreqArr = adjFreqArr;
      filterResponse.effectiveNoiseBandwidth = effectiveNoiseBandwidth;
      filterResponse.minus3DbCutoffFreq = minus3DbCutoffFreq;
      filterResponse.magnitudeArr = magnitudeArr;
      filterResponse.normalModeRangeDetails = normalModeRangeDetails;
      filterResponse.effectiveDataRate = effectiveDataRate;
      filterResponse.fclk_nom = fclk_nom;
      isFilterResponseCalculated = true;
    }
  } else {
    isFilterResponseCalculated = false;
  }
  let table = [];
  table.push(
    utility.getTableRow(
      [
        {
          value: "Digital Filter Options",
          colspan: 3,
        },
        {
          value: "Normal Mode Rejection",
        },
      ],
      "head",
    ),
  );

  table.push(
    utility.getTableRow(
      [
        {
          value: "Effective Data Rate",
        },
        {
          value: "-3 dB Cutoff Frequency",
        },
        {
          value: "Effective Noise Bandwidth",
        },
        {
          value: `Minimum Attenuation (${normalModeRangeDetails.min.toFixed(
            1,
          )} - ${normalModeRangeDetails.max.toFixed(1)} Hz)`,
        },
      ],
      "head",
    ),
  );

  table.push(
    utility.getTableRow([
      {
        value: `${effectiveDataRate.toFixed(2)} SPS`,
      },
      {
        value:
          minus3DbCutoffFreq !== undefined
            ? `${minus3DbCutoffFreq.toFixed(2)} Hz`
            : "N/A",
      },
      {
        value: `${effectiveNoiseBandwidth.toFixed(2)} Hz`,
      },
      {
        value: `${normalModeRangeDetails.minimumAttenuation.toFixed(2)} dB`,
      },
    ]),
  );

  return {
    value: table,
  };
}

export function computeTimeToConversionGraph(args) {
  let conversionTimings = [];
  const filter = args.filterType.value;
  const dataRateSetting = args.dataRateSetting.value;
  const chop =
    typeof args.globalChop.value === "boolean"
      ? Number(args.globalChop.value)
      : args.globalChop.value;
  const fClk_MHz = args.clockFrequencyMHz.value;
  const delay = args.programmableDelay.value;
  let module = args.parameterConfigData[args.selectedDevice].calculationModule;
  if (Object.keys(module).length > 0) {
    let timeTo1stConversion = module.getFilterSettlingTime(
      filter,
      dataRateSetting,
      fClk_MHz * 1000000,
      chop,
      delay,
    );
    let conversionPeriod = module.deviceConfig.single_conversion_time_ms;
    for (let n = 1; n <= 5; n++) {
      conversionTimings.push(timeTo1stConversion + (n - 1) * conversionPeriod);
    }
  } else {
    const programmableDelay = parseInt(
      args.delay_list.value[args.programmableDelay.value],
    );
    const chopEnabled = args.globalChop.value;
    const FIRST_DATA =
      args.first_data.value[args.filterType.value][dataRateSetting];
    const SECOND_DATA = args.data_rate_osr_value.value[dataRateSetting];

    for (let n = 1; n <= 5; n++) {
      conversionTimings.push(
        this.getTimeToNthConversion(
          fClk_MHz,
          n,
          programmableDelay,
          FIRST_DATA,
          SECOND_DATA,
          chopEnabled,
        ),
      );
    }
  }
  return {
    limits: {
      /**
       * In the graph, after the last conversion timing line, "..." will be shown to indicate the continuation.
       *
       * So, to accommodate that, step size of 1 is added. (conversionTimings.at(-1) + 1)
       */
      xaxisMax: conversionTimings.at(-1) + 1,
      xaxisMin: 0,
    },
    metaData: {
      conversionTimings,
      conversionTimingsUnit: "ms",
    },
  };
}

// ---------- Utilies ---------------

export function getAdjustedFreq(fClk_MHz, fClkNom_MHz, freq) {
  return (fClk_MHz / fClkNom_MHz) * freq;
}

export function getFilterResponseDetails(
  filterType,
  chopEnabled,
  sinc3_OSR,
  sinc1_OSR_List,
  dataRateSetting,
  lutData,
  fClk_MHz,
  fClkNom_MHz,
  frequencyInp,
  toleranceInp,
) {
  let adjFreqArr = [];
  let magnitudeArr = [];
  let effectiveNoiseBandwidth = 0;
  let minus3DbCutoffFreq;
  let normalModeRange = {
    availableFMin: undefined,
    availableFMax: undefined,
    calculatedFMin: frequencyInp - toleranceInp,
    calculatedFMax: frequencyInp + toleranceInp,
    minimumAttenuation: undefined,
  };
  let osrList = getOSRList(
    filterType,
    dataRateSetting,
    sinc3_OSR,
    sinc1_OSR_List,
  );
  let totalOSR = osrList.reduce(
    (accumulator, currentValue) => accumulator * currentValue,
  );

  let y = -3;
  let x0, x1;
  let y0 = Infinity;
  let y1 = -Infinity;
  let pointsFound = false;
  let freq = 0;
  let index = 0;

  while (index < lutData["Freq"].length - 1) {
    let adjFreq;
    let magnitude = 0;

    freq = lutData["Freq"][index + 1];

    adjFreq =
      filterType === "FIR" && dataRateSetting > 11
        ? (dataRateSetting === 12 ? 2 : 4) *
          this.getAdjustedFreq(fClk_MHz, fClkNom_MHz, freq)
        : this.getAdjustedFreq(fClk_MHz, fClkNom_MHz, freq);
    if (filterType === "SINC3") {
      let sinc3_fdata_in = this.getFMod(fClk_MHz * 1000000);
      let sinc3FilterResponse = this.getSinc3FilterResponse(
        adjFreq,
        sinc3_fdata_in,
        sinc3_OSR,
      );
      magnitude += this.getMagnitude(sinc3FilterResponse);
    } else if (dataRateSetting <= 2) {
      let sinc3_fdata_in = this.getFMod(fClk_MHz * 1000000);
      let sinc3FilterResponse = this.getSinc3FilterResponse(
        adjFreq,
        sinc3_fdata_in,
        320,
      );
      magnitude += this.getMagnitude(sinc3FilterResponse);
      let sinc1_fdata_in =
        this.getFMod(fClk_MHz * 1000000) / (osrList[0] * osrList[1]);
      let sinc1_OSR = sinc1_OSR_List[dataRateSetting];
      let sinc1FilterResponse = this.getSinc1FilterResponse(
        adjFreq,
        sinc1_fdata_in,
        sinc1_OSR,
      );
      magnitude += this.getMagnitude(sinc1FilterResponse);
      magnitude += lutData.D1[index];
    } else if (dataRateSetting === 4) {
      let sinc3_fdata_in = this.getFMod(fClk_MHz * 1000000);
      let sinc3FilterResponse = this.getSinc3FilterResponse(
        adjFreq,
        sinc3_fdata_in,
        320,
      );
      magnitude += this.getMagnitude(sinc3FilterResponse);
      magnitude += lutData.D1[index];
    } else if (dataRateSetting >= 11) {
      magnitude += lutData.A1[index];
    } else if (dataRateSetting === 6) {
      let sinc1_fdata_in = this.getFMod(fClk_MHz * 1000000) / osrList[0];
      let sinc1_OSR = sinc1_OSR_List[dataRateSetting];
      let sinc1FilterResponse = this.getSinc1FilterResponse(
        adjFreq,
        sinc1_fdata_in,
        sinc1_OSR,
      );
      magnitude += this.getMagnitude(sinc1FilterResponse);
      magnitude += lutData.C1[index];
    } else if (dataRateSetting === 10) {
      magnitude += lutData.B1[index];
    } else {
      let sinc1_fdata_in = this.getFMod(fClk_MHz * 1000000) / osrList[0];
      let sinc1_OSR = sinc1_OSR_List[dataRateSetting];
      let sinc1FilterResponse = this.getSinc1FilterResponse(
        adjFreq,
        sinc1_fdata_in,
        sinc1_OSR,
      );
      magnitude += this.getMagnitude(sinc1FilterResponse);
      magnitude += lutData.B1[index];
    }
    if (chopEnabled) {
      let chop_fdata_in =
        this.getFMod(fClk_MHz * 1000000) /
        osrList.reduce(
          (accumulator, currentValue) => accumulator * currentValue,
        );
      let chopFilterResponse = this.getChopFilterResponse(
        adjFreq,
        chop_fdata_in,
      );
      magnitude += this.getMagnitude(chopFilterResponse);
    }
    adjFreqArr.push(adjFreq);
    magnitudeArr.push(magnitude);

    // Effective noise bandwidth calculation
    if (index !== 0) {
      let changeInFreq = adjFreqArr[index] - adjFreqArr[index - 1];

      let yAverage =
        ((this.magnitudeToFilterResponse(magnitudeArr[index]) +
          this.magnitudeToFilterResponse(magnitudeArr[index - 1])) /
          2) **
        2;

      effectiveNoiseBandwidth += changeInFreq * yAverage;
    }

    // -3 dB cutoff frequency initial data
    if (magnitude < y && magnitude > y1 && !pointsFound) {
      y1 = magnitude;
      x1 = adjFreq;
    } else if (magnitude > y && magnitude < y0 && !pointsFound) {
      y0 = magnitude;
      x0 = adjFreq;
    } else if (y0 != Infinity && y1 != -Infinity) {
      pointsFound = true;
    }

    // Normal mode range (Min, Max)
    ["Min", "Max"].forEach(range => {
      if (normalModeRange[`availableF${range}`] === undefined) {
        if (adjFreq >= normalModeRange[`calculatedF${range}`]) {
          normalModeRange[`availableF${range}`] = {
            freq: adjFreq,
            index,
          };
        }
      }
    });

    // Normal mode range (minimum attenuation)
    if (
      // Once minimum frequency is found,
      normalModeRange.availableFMin !== undefined &&
      // Till maximum frequency is reached (inclusive of maximum frequency)
      (normalModeRange.availableFMax === undefined ||
        normalModeRange.availableFMax.index === index)
    ) {
      normalModeRange.minimumAttenuation =
        normalModeRange.minimumAttenuation === undefined ||
        normalModeRange.minimumAttenuation < magnitude
          ? magnitude
          : normalModeRange.minimumAttenuation;
    }

    index++;
  }

  // -3 dB cutoff frequency calculation
  minus3DbCutoffFreq = x0 + (y - y0) * ((x1 - x0) / (y1 - y0));

  // ---------- Incase all the frequencies are below calculatedFMin/Max ----------

  ["Min", "Max"].forEach(range => {
    if (normalModeRange[`availableF${range}`] === undefined) {
      normalModeRange[`availableF${range}`] = {
        freq: adjFreqArr[adjFreqArr.length - 1],
        index: adjFreqArr.length - 1,
      };
    }
  });

  let effectiveDataRate = this.getEffectiveDataRate(fClk_MHz, totalOSR);

  return {
    adjFreqArr,
    effectiveNoiseBandwidth,
    magnitudeArr,
    minus3DbCutoffFreq,
    normalModeRangeDetails: {
      min: normalModeRange.availableFMin.freq,
      max: normalModeRange.availableFMax.freq,
      minimumAttenuation:
        normalModeRange.minimumAttenuation !== undefined
          ? normalModeRange.minimumAttenuation
          : magnitudeArr[magnitudeArr.length - 1],
    },
    effectiveDataRate,
  };
}

export function getFilterResponseOutputs(
  adjFreqArr,
  magnitudeArr,
  frequencyInp,
  toleranceInp,
) {
  let y = -3;
  let x0, x1;
  let y0 = Infinity;
  let y1 = -Infinity;
  let pointsFound = false;
  let index = 0;
  let effectiveNoiseBandwidth = 0;
  let minus3DbCutoffFreq;
  let normalModeRange = {
    availableFMin: undefined,
    availableFMax: undefined,
    calculatedFMin: frequencyInp - toleranceInp,
    calculatedFMax: frequencyInp + toleranceInp,
    minimumAttenuation: undefined,
  };

  while (index < adjFreqArr.length - 1) {
    let magnitude = magnitudeArr[index];
    let adjFreq = adjFreqArr[index];
    // Effective noise bandwidth calculation
    if (index !== 0) {
      let changeInFreq = adjFreqArr[index] - adjFreqArr[index - 1];

      let yAverage =
        ((this.magnitudeToFilterResponse(magnitudeArr[index]) +
          this.magnitudeToFilterResponse(magnitudeArr[index - 1])) /
          2) **
        2;

      effectiveNoiseBandwidth += changeInFreq * yAverage;
    }

    // -3 dB cutoff frequency initial data
    if (magnitude < y && magnitude > y1 && !pointsFound) {
      y1 = magnitude;
      x1 = adjFreq;
    } else if (magnitude > y && magnitude < y0 && !pointsFound) {
      y0 = magnitude;
      x0 = adjFreq;
    } else if (y0 != Infinity && y1 != -Infinity) {
      pointsFound = true;
    }

    // Normal mode range (Min, Max)
    ["Min", "Max"].forEach(range => {
      if (normalModeRange[`availableF${range}`] === undefined) {
        if (adjFreq >= normalModeRange[`calculatedF${range}`]) {
          normalModeRange[`availableF${range}`] = {
            freq: adjFreq,
            index,
          };
        }
      }
    });

    // Normal mode range (minimum attenuation)
    if (
      // Once minimum frequency is found,
      normalModeRange.availableFMin !== undefined &&
      // Till maximum frequency is reached (inclusive of maximum frequency)
      (normalModeRange.availableFMax === undefined ||
        normalModeRange.availableFMax.index === index)
    ) {
      normalModeRange.minimumAttenuation =
        normalModeRange.minimumAttenuation === undefined ||
        normalModeRange.minimumAttenuation < magnitude
          ? magnitude
          : normalModeRange.minimumAttenuation;
    }

    index++;
  }

  // -3 dB cutoff frequency calculation
  minus3DbCutoffFreq = x0 + (y - y0) * ((x1 - x0) / (y1 - y0));

  // ---------- Incase all the frequencies are below calculatedFMin/Max ----------

  ["Min", "Max"].forEach(range => {
    if (normalModeRange[`availableF${range}`] === undefined) {
      normalModeRange[`availableF${range}`] = {
        freq: adjFreqArr[adjFreqArr.length - 1],
        index: adjFreqArr.length - 1,
      };
    }
  });

  return {
    effectiveNoiseBandwidth,
    minus3DbCutoffFreq,
    normalModeRangeDetails: {
      min: normalModeRange.availableFMin.freq,
      max: normalModeRange.availableFMax.freq,
      minimumAttenuation:
        normalModeRange.minimumAttenuation !== undefined
          ? normalModeRange.minimumAttenuation
          : magnitudeArr[magnitudeArr.length - 1],
    },
  };
}

export function getOSRList(
  filterType,
  dataRateSetting,
  sinc3_OSR,
  sinc1_OSR_List,
) {
  let osrList = [];
  if (filterType === "SINC3" || dataRateSetting >= 11) {
    osrList.push(sinc3_OSR);
    return osrList;
  } else if (dataRateSetting <= 2) {
    let sinc3_OSR = 320;
    let d1_OSR = 40;
    osrList.push(sinc3_OSR, d1_OSR, sinc1_OSR_List[dataRateSetting]);
    return osrList;
  } else if (dataRateSetting === 4) {
    let sinc3_OSR = 320;
    let d1_OSR = 40;
    osrList.push(sinc3_OSR, d1_OSR);
    return osrList;
  } else if (dataRateSetting === 6) {
    let c1_OSR = 328;
    osrList.push(c1_OSR, sinc1_OSR_List[dataRateSetting]);
    return osrList;
  } else if (dataRateSetting === 10) {
    let b1_OSR = 320;
    osrList.push(b1_OSR);
    return osrList;
  } else {
    let b1_OSR = 320;
    osrList.push(b1_OSR, sinc1_OSR_List[dataRateSetting]);
    return osrList;
  }
}

export function getEffectiveDataRate(fClk_MHz, OSR) {
  return (fClk_MHz * 1000000) / (16 * OSR);
}

export function getFMod(fClk_MHz) {
  return fClk_MHz / 16;
}

export function getFreqFromAdjustedFreqFunc(fClk_MHz, fClkNom_MHz) {
  return function (adjFreq) {
    return adjFreq / (fClk_MHz / fClkNom_MHz);
  };
}

export function getMagnitude(sinc3FilterResponse) {
  return 20 * Math.log10(sinc3FilterResponse);
}

export function getSinc3FilterResponse(adjFreq, fdata_in, OSR) {
  return (
    Math.abs(
      Math.sin((Math.PI * adjFreq * OSR) / fdata_in) /
        (OSR * Math.sin((Math.PI * adjFreq) / fdata_in)),
    ) ** 3
  );
}

export function getSinc1FilterResponse(adjFreq, fdata_in, OSR) {
  return Math.abs(
    Math.sin((Math.PI * adjFreq * OSR) / fdata_in) /
      (OSR * Math.sin((Math.PI * adjFreq) / fdata_in)),
  );
}

export function getChopFilterResponse(adjFreq, fdata_in) {
  return Math.abs(
    Math.sin((Math.PI * adjFreq * 2) / fdata_in) /
      (2 * Math.sin((Math.PI * adjFreq) / fdata_in)),
  );
}

export function getTimeTo1stConversion(
  fClk_MHz,
  programmableDelay,
  FIRST_DATA,
  chopEnabled,
) {
  let timeTo1stConversion =
    ((FIRST_DATA + programmableDelay) * this.getTMod(fClk_MHz)) / 1000;
  return chopEnabled ? timeTo1stConversion * 2 : timeTo1stConversion;
}

export function getConversionPeriod(
  fClk_MHz,
  OSR,
  timeTo1stConversion,
  chopEnabled,
) {
  if (chopEnabled) {
    return timeTo1stConversion / 2;
  } else {
    return 1000 / this.getEffectiveDataRate(fClk_MHz, OSR);
  }
}

export function getTimeToNthConversion(
  fClk_MHz,
  n,
  programmableDelay,
  SINC3_FIRST_DATA,
  SECOND_DATA,
  chopEnabled,
) {
  let timeTo1stConversion = this.getTimeTo1stConversion(
    fClk_MHz,
    programmableDelay,
    SINC3_FIRST_DATA,
    chopEnabled,
  );
  return chopEnabled
    ? timeTo1stConversion + (n - 1) * (timeTo1stConversion / 2)
    : timeTo1stConversion +
        (n - 1) * ((SECOND_DATA * this.getTMod(fClk_MHz)) / 1000);
}

export function getTMod(fClk_MHz) {
  return 16 / fClk_MHz;
}

export function magnitudeToFilterResponse(magnitude) {
  return 10 ** (magnitude / 20);
}
